@@ -14,22 +14,42 @@ from trace_view import view_trace
 import sys
 import glob
 import numpy as np
-from stitch import stitch
+import stitch
 import gc
 from collections import defaultdict
+import service
 
-ATT_VERSION = 2
+ATT_VERSION = 5
+
+class pcInfo(ctypes.Structure):
+    _fields_ = [
+        ("addr", c_uint64),
+        ("marker_id", c_uint64)
+    ]
+    def to_v2_pc(self):
+        if self.marker_id == 0:
+            return self.addr
+        return self.addr | (self.marker_id<<service.ID_OFFSET) | (1<<service.HEADER_OFFSET)
 
 class TraceData(ctypes.Structure):
     _fields_ = [
-        ("num_waves", c_uint64),
-        ("type", c_uint64),
-        ("cycles", c_uint64)
+        ("type", c_uint64, 8),
+        ("hitcount", c_uint64, 56),
+        ("latency", c_uint64),
+        ("pc", pcInfo)
     ]
 
+class TraceDataTranslated:
+    def __init__(self, inst):
+        self.type = inst.type
+        self.num_waves = inst.hitcount
+        self.cycles = inst.latency
+        if self.type == stitch.PCINFO:
+            self.cycles = inst.pc.to_v2_pc()
+
 class Trace:
     def __init__(self, traceid, tracesize, instructions_array):
-        self.instructions = [deepcopy(instructions_array[k]) for k in range(tracesize)]
+        self.instructions = [TraceDataTranslated(instructions_array[k]) for k in range(tracesize)]
         self.traceid = traceid
 
 class PerfEvent(ctypes.Structure):
@@ -80,43 +100,61 @@ class ReturnAssemblyInfo(ctypes.Structure):
                 ('jumps_len', ctypes.c_int)]
 
 
+class WaveState(ctypes.Structure):
+    _fields_ = [
+        ("type", c_int32),
+        ("duration", c_int32)
+    ]
+
+class WaveInstruction(ctypes.Structure):
+    _fields_ = [
+        ("time", c_int64),
+        ("duration", c_int64)
+    ]
+
 class Wave(ctypes.Structure):
     _fields_ = [
-        ("simd", ctypes.c_uint64),
-        ("wave_id", ctypes.c_uint64),
-        ("begin_time", ctypes.c_uint64),  # Begin and end cycle
-        ("end_time", ctypes.c_uint64),
-        ("traceid", ctypes.c_int64),
-        # total VMEM/FLAT/LDS/SMEM instructions issued
-        # total issued memory instructions
-        ("num_mem_instrs", ctypes.c_uint64),
-        # total issued instructions (compute + memory)
-        ("num_issued_instrs", ctypes.c_uint64),
-        ("num_valu_instrs", ctypes.c_uint64),
-        ("num_valu_stalls", ctypes.c_uint64),
-        # VMEM Pipeline: instrs and stalls
-        ("num_vmem_instrs", ctypes.c_uint64),
-        ("num_vmem_stalls", ctypes.c_uint64),
-        # FLAT instrs and stalls
-        ("num_flat_instrs", ctypes.c_uint64),
-        ("num_flat_stalls", ctypes.c_uint64),
-        # LDS instr and stalls
-        ("num_lds_instrs", ctypes.c_uint64),
-        ("num_lds_stalls", ctypes.c_uint64),
-        # SCA instrs stalls
-        ("num_salu_instrs", ctypes.c_uint64),
-        ("num_smem_instrs", ctypes.c_uint64),
-        ("num_salu_stalls", ctypes.c_uint64),
-        ("num_smem_stalls", ctypes.c_uint64),
-        # Branch
-        ("num_branch_instrs", ctypes.c_uint64),
-        ("num_branch_taken_instrs", ctypes.c_uint64),
-        ("num_branch_stalls", ctypes.c_uint64),
+        ("simd", ctypes.c_uint8),
+        ("wave_id", ctypes.c_uint8),
+        ("trap_status", ctypes.c_uint8),
+        ("reserved", ctypes.c_uint8),
 
-        ("timeline_size", ctypes.c_uint64),
-        ("instructions_size", ctypes.c_uint64),
-        ("timeline_array", POINTER(ctypes.c_int32)),
-        ("instructions_array", POINTER(ctypes.c_uint64)),
+        # total VMEM/FLAT/LDS/SMEM instructions issued
+        # VMEM Pipeline: instrs and stalls
+        ("num_vmem_instrs", ctypes.c_int),
+        ("num_vmem_stalls", ctypes.c_int),
+        # FLAT instrs and stalls
+        ("num_flat_instrs", ctypes.c_int),
+        ("num_flat_stalls", ctypes.c_int),
+        # LDS instr and stalls
+        ("num_lds_instrs", ctypes.c_int),
+        ("num_lds_stalls", ctypes.c_int),
+        # SCA instrs stalls
+        ("num_salu_instrs", ctypes.c_int),
+        ("num_smem_instrs", ctypes.c_int),
+        ("num_salu_stalls", ctypes.c_int),
+        ("num_smem_stalls", ctypes.c_int),
+        # Branch
+        ("num_branch_instrs", ctypes.c_int),
+        ("num_branch_taken_instrs", ctypes.c_int),
+        ("num_branch_stalls", ctypes.c_int),
+
+        # total issued memory instructions
+        ("num_mem_instrs", ctypes.c_int),
+        # total valus insts and stalls
+        ("num_valu_stalls", ctypes.c_int),
+        ("num_valu_instrs", ctypes.c_size_t),
+        # total issued instructions (compute + memory)
+        ("num_issued_instrs", ctypes.c_size_t),
+        # Begin and end cycle
+        ("begin_time", ctypes.c_int64),
+        ("end_time", ctypes.c_int64),
+        ("traceid", ctypes.c_int64),
+
+        ("timeline_size", ctypes.c_size_t),
+        ("instructions_size", ctypes.c_size_t),
+        ("timeline_array", POINTER(WaveState)),
+        ("instructions_array", POINTER(WaveInstruction)),
     ]
 
 
@@ -129,13 +167,13 @@ class PythonWave:
                 pass
 
         self.timeline = [
-            (int(sourcew.timeline_array[2 * k]), int(sourcew.timeline_array[2 * k + 1]))
+            (int(sourcew.timeline_array[k].type), int(sourcew.timeline_array[k].duration))
             for k in range(self.timeline_size)
         ]
         self.timeline_array = None
 
         self.instructions = [
-            (int(sourcew.instructions_array[2*k+0]), int(sourcew.instructions_array[2*k+1]))
+            (int(sourcew.instructions_array[k].time), int(sourcew.instructions_array[k].duration))
             for k in range(self.instructions_size)
         ]
         self.instructions_array = None
@@ -154,7 +192,7 @@ class ReturnInfo(ctypes.Structure):
         ("perfevents", POINTER(PerfEvent)),
         ("occupancy", POINTER(ctypes.c_uint64)),
         ("num_occupancy", ctypes.c_uint64),
-        ("kernel_id_addr", POINTER(ctypes.c_uint64)),
+        ("kernel_id_addr", POINTER(pcInfo)),
         ("num_kernel_ids", ctypes.c_uint64),
 
         ("wavedata", POINTER(Wave)),
@@ -189,15 +227,10 @@ def parse_binary(filename, kernel=None):
     info = SO.wrapped_parse_binary(str(filename).encode("utf-8"), kernel)
 
     code = []
-    kernel_addr = defaultdict(lambda : "Unknown")
-    last_known_function = "Unknown"
     for k in range(info.code_len):
         code_entry = info.code[k]
 
         line = deepcopy(code_entry.line.decode("utf-8"))
-        if "; Begin " in line:
-            last_known_function = line.split("; Begin ")[1]
-
         loc = deepcopy(code_entry.loc.decode("utf-8"))
 
         to_line = int(code_entry.to_line) if (code_entry.to_line >= 0) else None
@@ -207,14 +240,11 @@ def parse_binary(filename, kernel=None):
         code.append([line, int(code_entry.value), to_line, loc, int(code_entry.index),
                     int(code_entry.line_num), int(code_entry.addr), 0, 0])
 
-        if code[-1][-3] != 0 and len(code) > 1:
-            kernel_addr[code[-1][-3]] = last_known_function
-
     jumps = {}
     for k in range(info.jumps_len):
         jumps[info.jumps[k].key] = info.jumps[k].value
 
-    return code, jumps, kernel_addr
+    return code, jumps
 
 
 def getWaves_binary(name):
@@ -226,7 +256,7 @@ def getWaves_binary(name):
         print('Invalid trace ', name)
         return ([], [], [], [], None, [])
     flags = "navi" if (info.flags & 0x2) else "vega"
-    kernel_addr = [int(info.kernel_id_addr[k]) for k in range(info.num_kernel_ids)]
+    kernel_addr = [info.kernel_id_addr[k].to_v2_pc() for k in range(info.num_kernel_ids)]
     events = [deepcopy(info.perfevents[k]) for k in range(info.num_events)]
     occupancy = [int(info.occupancy[k]) for k in range(int(info.num_occupancy))]
 
@@ -245,16 +275,20 @@ def getWaves_binary(name):
 
     SO.FreeBinary(info.binaryID)
 
-    return (traces_python, waves_python, events, occupancy, flags, kernel_addr)
+    return (traces_python, waves_python, events, occupancy, flags, kernel_addr, info.flags & 0x4)
 
 
-def getWaves_stitch(traces, code, jumps, flags, latency_map, hitcount_map, bIsAuto):
+def getWaves_stitch(traces, code, jumps, flags, latency_map, hitcount_map, bIsAuto, codeobjservice):
     for id in traces.keys():
-        traces[id].instructions = stitch(traces[id].instructions, code, jumps, flags, bIsAuto)
+        traces[id].instructions = stitch.stitch(traces[id].instructions, code, jumps, flags, bIsAuto, codeobjservice)
+        if len(code) > hitcount_map.size:
+            hitcount_map = np.pad(hitcount_map, [0,len(code)-hitcount_map.size])
+            latency_map = np.pad(latency_map, [0,len(code)-latency_map.size])
         if traces[id].instructions is not None:
             for inst in traces[id].instructions[0]:
                 hitcount_map[inst.asmline] += inst.num_waves
                 latency_map[inst.asmline] += inst.cycles
+    return hitcount_map, latency_map
 
 
 def persist(trace_file, SIMD, traces):
@@ -296,12 +330,12 @@ def persist(trace_file, SIMD, traces):
         br_stalls.append(wave.num_branch_stalls)
         timeline.append(wave.timeline)
 
-        cc = 1
+        cc = 0
         insts = []
         skips = traces[wave.traceid].instructions[-1]
         try:
             for v in traces[wave.traceid].instructions[0]:
-                if cc in skips:
+                while cc in skips:
                     cc += 1
                 t = wave.instructions[cc]
                 insts.append((t[0], v.type, 0, t[1], v.asmline))
@@ -434,28 +468,28 @@ if __name__ == "__main__":
         "--trace_file", help="Filter for trace files", default=None, type=str
     )
     parser.add_argument(
-        "--att_kernel", help="Kernel file", type=str, default=pathenv + "/*_kernel.txt"
+        "--att_kernel", help="Kernel file", type=str, default=os.path.join(pathenv, "*_kernel.txt")
     )
     parser.add_argument("--ports", help="Server and websocket ports, default: 8000,18000")
     parser.add_argument(
         "--mode",
         help="""ATT analysis modes:\n
                         off: Only run ATT collection, disable analysis.\n
-                        file: dump json files to disk.\n
-                        network: Open att server over the network.""",
+                        file: dump json files to disk.""",
         type=str,
         default="off",
     )
     args = parser.parse_args()
+    args.mode = args.mode.lower().split(',')
 
     CSV_MODE = False
-    if args.mode.lower() == 'csv':
+    FILE_MODE = False
+    if 'csv' in args.mode:
         CSV_MODE = True
-    elif args.mode.lower() == 'file':
-        args.dumpfiles = True
-    elif args.mode.lower() == "network":
-        args.dumpfiles = False
-    else:
+    if 'file' in args.mode:
+        FILE_MODE = True
+
+    if not CSV_MODE and not FILE_MODE:
         print("Skipping analysis.")
         quit()
 
@@ -480,17 +514,39 @@ if __name__ == "__main__":
 
     for att_kernel in att_kernel_list:
         print('Parsing:', att_kernel)
+        att_kernel_f = []
+        with open(att_kernel, 'r') as f:
+            for line in f:
+                att_kernel_f.append(line.split('\n')[0])
+
+        # returns element index of the file:// or memory:// filepath in the string
+        get_path_loc = lambda x: np.sum([len(m) for m in x.split(' ')[:3]])+3
+        # adds the OUTPUT_PATH env variable to a filepath if necessary
+        add_pathnv = lambda x: x[:get_path_loc(x)] + os.path.join(pathenv, x[get_path_loc(x):])
+        # returns the memory address in the string
+        get_addr = lambda x: int(x.split(' ')[0][2:], 16)
+        # Sets a preference for 'file' paths to be added before 'memory' paths. Sorts by addr.
+        get_addr_preference = lambda x: [0 if 'file' in x[get_path_loc(x):] else 1<<60][0]
+
+        # Get the GPU id in the string
+        gpu_id = int(att_kernel_f[0].split(' ')[2].split('GPU[')[1].split(']')[0])
+        # Eliminame first line in the att_kernel txt file and adds the OUTPUT_PATH as needed
+        att_kernel_f = [add_pathnv(p) if '.out' == p[-4:] else p for p in att_kernel_f[1:]]
+        # Sorts the list of codeobj by address, with 'file' given preference
+        att_kernel_f = sorted(att_kernel_f, key=lambda x: get_addr(x)+get_addr_preference(x))
         assembly_code = deepcopy(args.assembly_code)
 
         # Assembly parsing
         bIsAuto = False
         if assembly_code.lower().strip() == 'auto':
-            assembly_code = att_kernel.split('_kernel.txt')[0]+'_isa.s'
             bIsAuto = True
-        path = Path(assembly_code)
-        if not path.is_file():
+            code = [['; Begin ATT ASM', 100, 0, '', 0, 0, 0, 0, 0]]
+            jumps = []
+        elif not Path(assembly_code).is_file():
             print("Invalid assembly_code('{0}')!".format(assembly_code))
             sys.exit(1)
+        else:
+            code, jumps = parse_binary(assembly_code, att_kernel)
 
         # Trace Parsing
         trace_instance_name = att_kernel.split("_kernel.txt")[0]
@@ -504,7 +560,6 @@ if __name__ == "__main__":
             continue
 
         print('Att kernel:', att_kernel)
-        code, jumps, kern_addr = parse_binary(assembly_code, None if bIsAuto else att_kernel)
 
         DBFILES = []
         EVENTS = []
@@ -512,21 +567,27 @@ if __name__ == "__main__":
         GFXV = []
         analysed_filenames = []
         occupancy_filenames = []
-        dispatch_kernel_names = {}
+        kernel_addr = {}
 
         latency_map = np.zeros((len(code)), dtype=np.int64)
         hitcount_map = np.zeros((len(code)), dtype=np.int32)
 
         gc.collect()
+        if bIsAuto:
+            codeservice = service.CodeobjService(gpu_id, att_kernel_f, SO.classify_asm_line)
+        else:
+            codeservice = None
 
         for name in filenames:
-            traces, waves, perfevents, occupancy, gfxv, addrs = getWaves_binary(name)
+            traces, waves, perfevents, occupancy, gfxv, addrs, ftrace = getWaves_binary(name)
             if gfxv is None:
                 continue
+            if CSV_MODE == False and ftrace == 0:
+                print('Generating occupancy information')
 
             for id, addr in enumerate(addrs):
-                dispatch_kernel_names[id] = kern_addr[addr]
-            if len(occupancy) > 16:
+                kernel_addr[id] = addr
+            if len(occupancy) > 1:
                 OCCUPANCY.append( occupancy )
                 occupancy_filenames.append(name)
 
@@ -534,7 +595,7 @@ if __name__ == "__main__":
                 print("No traces from", name)
                 continue
 
-            getWaves_stitch(traces, code, jumps, gfxv, latency_map, hitcount_map, bIsAuto)
+            hitcount_map, latency_map = getWaves_stitch(traces, code, jumps, gfxv, latency_map, hitcount_map, bIsAuto, codeservice)
 
             analysed_filenames.append(name)
             EVENTS.append(perfevents)
@@ -545,25 +606,34 @@ if __name__ == "__main__":
         for k in range(len(code)):
             code[k][-2] = int(hitcount_map[k])
             code[k][-1] = int(latency_map[k])
+            try:
+                code[k][-3] = codeservice.ToRawPC(code[k][-3])
+            except:
+                pass
+            if k > 0 and code[k-1][-3] == 0:
+                code[k-1][-3] = code[k][-3]
 
         if CSV_MODE:
             from att_to_csv import dump_csv
-            dump_csv(code, trace_instance_name)
-        else:
+            dump_csv(code, trace_instance_name, bIsAuto)
+
+        if FILE_MODE:
+            try:
+                dispatchNames = {id: codeservice.getSymbolName(addr) for id, addr in kernel_addr.items()}
+            except:
+                dispatchNames = {id: "#addr"+str(addr) for id, addr in kernel_addr.items()}
             drawinfo = {
                 "TIMELINES": gen_timelines(DBFILES),
                 "EVENTS": EVENTS,
                 "EVENT_NAMES": EVENT_NAMES,
                 "OCCUPANCY": OCCUPANCY,
                 "ShaderNames": occupancy_filenames,
-                "DispatchNames": dispatch_kernel_names,
+                "DispatchNames": dispatchNames
             }
             view_trace(
-                args,
                 code,
                 DBFILES,
                 analysed_filenames,
-                args.dumpfiles,
                 0,
                 gfxv,
                 drawinfo,
